// --- [ sun.c ] --------------------------------------------------------------
//       01.05.09
// Der Rechenweg fr den Sonnenstand kommt von:
//-----------------------------------------------------------------------------
// Prozessor : ATmega8
// Sprache   : C
// Datum     : 20.09.2008
// Version   : 1.0
// Autor     : Cusoe
// Programm  : vereinfachte Sonnenstandsberechnung
//-----------------------------------------------------------------------------

// die Berechnung des Sonnenstandes ist bernommen von: siehe oben

#include	<avr\io.h>				//AVR Register und Konstantendefinitionen
#include	<avr\interrupt.h> 	//AVR Interrupt Vektoren
#include	<stdlib.h>
#include	<inttypes.h>
#include	<math.h>
#include "global.h"
#include "lcd.h"
#include "sun.h"

#ifdef SUN_INCLUDE

const double geo_laenge = 10.50;			// Geographische Lnge  des Standortes: 10'30"
const double geo_breite = 52.15;			// Geographische Breite des Standortes: 52'10"
													// Annahme: Ortszeit = UTC + 1h
double azimut, elevation;

// ----------------------------------------------------------------------------
// Berechnet und zeigt aktuelle Elevation, Azimut sowie Sonnenauf-/Untergangszeit
// stellt zuerst das Vorzeichen fest, wandelt ggf. in positiven Wert um
// multipliziert mit 100, damit 2 Nachkommastellen mitgenommen werden
// wandelt dann die Floats um in uint16_t
// und legt upper und lower Byte in Array SUN_data ab.
// und zeigt natrlich die Werte auch noch an.

void SUN_display (void)
{
static uint8_t x = 0;						// signalisiert, ob der Sonnenauf-/Untergang zu berechnen sind
uint16_t  tmp;
int16_t i_tmp;
uint8_t  sign;

DCF_make_UTC_time();							// UTC-Zeit aus der aktuellen Uhrzeit erzeugen

SUN_calc();										// Elevation und Azimut fr UTC-Zeit berechnen 
													// und in die globalen Variablen 'elevation' schreiben

i_tmp = (int16_t) (elevation * 100);	// fr zwei Nachkommastellen mit 100 multiplizieren
time[SUN][SUN_EL_HIGH] = (uint8_t) (i_tmp / 256); // zerlegen und als int16_t speichern
time[SUN][SUN_EL_LOW]  = (uint8_t)  i_tmp;

if (i_tmp < 0)
	{
	tmp = (uint16_t) (0 - i_tmp);			// zur Ausgabe umwandeln in uint16_t (nach tmp)
	sign = MINUS;
	}
else
	{
	tmp = (uint16_t) (i_tmp);
	sign = PLUS;
	}

#ifdef LCD
lcd_goto_xy(3,6);								// Elevation +- 0 ... 90
SUN_print(tmp, sign, 2);					// Ausgeben mit Vorzeichen +/-
#endif

tmp = (uint16_t) (azimut * 100);		// Azimut 0 ... 360
time[SUN][SUN_AZ_HIGH] = (uint8_t) (tmp / 256); // zerlegen und speichern
time[SUN][SUN_AZ_LOW]  = (uint8_t)  tmp;

#ifdef LCD
lcd_goto_xy(4,5);	
SUN_print(tmp, ' ', 3);						// kein Vorzeichen ausgeben
#endif
													// Direkt nach dem Programmstart ist x = 0.
													// der Tag ist aber immer 1 oder grer
													// nach jeder nderung des Datum oder der Sommer/Winterzeit
if (!(x == (time[CLK][TAG]) + is_summertime(&time[CLK][0])))	
	{												// wird der Sonnenauf- Untergang neu berechnen
													// Summe aus Tag + Sommerzeit als Merkmal
													// damit nur nach relevanten nderungen Neuberechnung erfolgt 
	x = time[CLK][TAG] + is_summertime(&time[CLK][0]);					

	SUN_check_sunrise();						// Durch Iteration Sonnenaufgang/Untergang bestimmen
													// das ist ein lngerer Rechenprozess
	}

#ifdef LCD
lcd_goto_xy(3, 13);							// Zeit fr Sonnenaufgang ausdrucken
lcd_dezprint(time[SUN][SUN_UP_H]);		// hours
lcd_putc(':');
lcd_dezprint(time[SUN][SUN_UP_M]);		// minutes

lcd_goto_xy(4, 13);							// Zeit fr Sonnenuntergang ausdrucken
lcd_dezprint(time[SUN][SUN_DN_H]);		// hours
lcd_putc(':');
lcd_dezprint(time[SUN][SUN_DN_M]);		// minutes
#endif

}

//----------------------------------------------------------------------
// Dezimaltag vom aktuellen Tag berechnen
//----------------------------------------------------------------------

float dezimal_day(void)
{
// wenn nur zu voller Minuten berechnet wird sollte dies ausreichen:
// die Sekundenbruchteile werden vernachlssigt
return(time[UTC][STUNDE] / 24.0 + time[UTC][MINUTE] / 1440.0 );
}

//----------------------------------------------------------------------
// Anzahl Dezimaltage seit 2000 bis zum aktuellen Datum berechnen
//----------------------------------------------------------------------

float julian_day(void)
{
uint8_t j;
j = time[UTC][JAHR] % 100;

// lfd. Nr. des Tage + Anzahl der Jahre ab 2000 * 365 ....
return((float) (day_of_year(&time[UTC][0]) + j * 365 + (j / 4) - 0.5));
}

//----------------------------------------------------------------------
// Reduzierung eines Winkel auf 0... 2PI rad, Stunden auf 0... 24 h etc.
// x=Wert der reduziert werden soll; y=Reduzierbereich z.B. auf 24h
//----------------------------------------------------------------------

float degr(float x, float y)
{

if(x < 0) x = x + y;
return((x - (floor(x / y) * y)));
}

// --------------------------------------------------------------------
// Den Sonnenauf/untergang durch "Probieren" bestimmen
// Rechnet ber Probieren:
// von 12.00h in 1 Stunden Intervallen die Zeit zurckzhlen bis  Elevation < x Grad,
// dann in 1 Minuten Intervallen vorwrts bis Elevation > -x Grad

// von 12.00h in Stunden-Intervallen die Zeit weiterzhlen bis Elevation < x Grad,
// dann in 1 Minuten-Intervallen zurck bis > x Grad

void SUN_check_sunrise(void)
{
// die UTC-Zeit wird zum Iterieren genutzt,
// sie muss am Ende der Funktion neu aufgebaut werden !
time[UTC][MINUTE] = 0;
time[UTC][STUNDE] = 12;

do	{ 												// Sonnenaufgang bestimmen
	time[UTC][STUNDE]--;						// um eine Stunde decrementieren
	SUN_calc();									// elevation neu berechnen
	} while (elevation > (ELEVATION));	// bis der gewhlte Grenzwert unterschritten ist

do {
	DCF_incr_time(&time[UTC][0], 1); 	// um 1 Minute incrementieren
	SUN_calc();									// elevation neu berechnen								
	} while (elevation < (ELEVATION));	// bis der gewhlte Grenzwert berschritten ist

// Fr unsere Zeitzone zur Winterzeit +1h und zur Sommerzeit +2h
time[SUN][SUN_UP_H] = ++time[UTC][STUNDE];
if (is_summertime(&time[CLK][0])) time[SUN][SUN_UP_H]++;
time[SUN][SUN_UP_M] = time[UTC][MINUTE];	// die Uhrzeit des Sonnenaufgangs speichern

time[UTC][MINUTE] = 0;
time[UTC][STUNDE] = 12;

do	{												// Sonnenuntergang bestimmen
	time[UTC][STUNDE]++;						// um eine Stunde incrementieren
	SUN_calc();									// elevation neu berechnen
	}
	while (elevation > (ELEVATION));		// bis der gewhlte Grenzwert unterschritten ist

do	{
	DCF_decr_time(&time[UTC][0], 1);		// um 1 Minute decrementieren
	SUN_calc();									// elevation neu berechnen
	} while (elevation < (ELEVATION));	// bis der gewhlte Grenzwert berschritten ist

// Fr unsere Zeitzone zur Winterzeit +1h und zur Sommerzeit +2h
time[SUN][SUN_DN_H] = ++time[UTC][STUNDE];
if (is_summertime(&time[CLK][0])) time[SUN][SUN_DN_H]++;	
time[SUN][SUN_DN_M] = time[UTC][MINUTE]; // die Uhrzeit des Sonnenuntergangs speichern

//DCF_make_UTC_time();							// UTC_Zeit neu aus der Uhrzeit generieren
}

//----------------------------------------------------------------------
// Vereinfachte Sonnenstandsberechnung
//----------------------------------------------------------------------
void SUN_calc(void)
{
double j, l, g, a, e, rekt, dekl, t0, sh, t, mz, rz, el, kr;
double jd0, dt, la, br;
double azi, elev;

jd0 = julian_day();
dt  = dezimal_day();

la  = geo_laenge / KO; 
br  = geo_breite / KO;

j = jd0 + dt;
l = degr(4.894950420 + 0.01720279239 * j, M_2PI); 	// Postion der Sonne auf der Ekliptik
g = degr(6.240040768 + 0.01720197034 * j, M_2PI); 	// mittlere Anomalie
a = l + 0.03342305518 * sin(g) + 0.0003490658504 * sin(2 * g); // ekliptikale Lnge der Sonne
e = 0.4090928042 - 0.000000006981317008 * j; 		// Schiefe der Ekliptik

rekt = atan(cos(e) * sin(a) / cos(a)); 				// Rektaszension

if(cos(a) < 0) rekt = rekt + M_PI;

dekl = asin(sin(e) * sin(a)); 							// Deklination

t0 = jd0 / 36525.0; 											// Tageszahl in julianischen Jahrhunderten
sh = degr(6.697376 + 2400.05134 * t0 + 1.002738 * dt * 24, 24); // mittlere Sternzeit
t = sh * 0.2617993878 + la - rekt; 						// Stundenwinkel der Sonne bezogen auf den Ort

mz = cos(t) * sin(br) - tan(dekl) * cos(br); 		// Nenner Azimut
rz = atan(sin(t) / mz) + M_PI; 							// Azimut wird von 0 Nord gezhlt

if(mz < 0) rz = rz + M_PI;									// wenn Nenner vom Azimut kleiner 0
																	// dann den Winkel in den richtigen Quadranten bringen
azi = degr(rz, M_2PI); 										// Azimut reduzieren
																	// Elevation
el = asin(cos(dekl) * cos(t) * cos(br) + sin(dekl) * sin(br));
																	// Korrekturwert fr Refraktion
kr = 0.0002967059728 / tan(el + 0.179768913 / (el + 0.08918632478));
elev = el + kr; 												// Elevation einschliesslich Refraktion

// das Ergebnisses in globale Variable schreiben

azimut    = (azi * KO);
elevation = (elev * KO);
}


// ---------------------------------------------------------------------------
// eine bereits 'vorverdaute' Float-Zahl ausgeben
#ifdef LCD
void SUN_print(uint16_t j, uint8_t c, uint8_t vorkomma)
{
register uint8_t i = 0;

lcd_putc(c);									// Vorzeichen ausgeben

if (vorkomma > 2)								// bei dreistelligen Ziffern
	{
	i = 0;
	while (j > 9999)
		{
		j -= 10000;
		i++;
		}
	lcd_putc('0' + i);						// Hunderterstelle
	}
	
i = 0;
while (j > 999)
	{
	j -= 1000;
	i++;
	}
lcd_putc('0' + i);							// Zehnerstelle

i = 0;
while (j > 99)
	{
	j -= 100;
	i++;
	}
lcd_putc('0' + i);							// Einerstelle

lcd_putc('.');									// Dezimalpunkt einschmuggeln

i = 0;
while (j > 9)
	{
	j -= 10;
	i++;
	}
lcd_putc('0' + i);							// 1. Nachkommastelle
lcd_putc('0' + j);							// 2. Nachkommastelle
}
#endif


#endif

// --- [ eof ] ----------------------------------------------------------------